;;       
;; 
;; Procedures for querying ion channel descriptions in NEMO models. 
;;
;; Copyright 2008-2011 Ivan Raikov and the Okinawa Institute of Science and Technology
;;
;; This program is free software: you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.
;;
;; A full copy of the GPL license can be found at
;; <http://www.gnu.org/licenses/>.
;;

(module nemo-gate-complex

 (nemo:ion-pool-query
  nemo:gate-complex-query)

 (import scheme chicken srfi-1 srfi-13)

 (require-extension environments matchable nemo-core nemo-utils)

(define (cid x)  (second x))
(define (cn x)   (first x))


(define (ispool? x)
  (match x (('decaying-pool name id) id) (('decaying 'pool name id) id) 
	 (else #f)))

(define (nemo:ion-pool-query sys)
   (match-let ((($ nemo:quantity 'DISPATCH  dis) (environment-ref sys (nemo-intern 'dispatch))))
     (let recur ((comp-name (nemo-intern 'toplevel)) (ax (list)))
       (let ((subcomps  ((dis 'component-subcomps)  sys comp-name)))
	 (let-values (((epool-comps other-comps)  (partition ispool? subcomps)))
		     (let ((epools (map (lambda (x) 
					  (let ((exports ((dis 'component-exports) sys (third x))))
					    (if (null? exports)
						(nemo:error 'nemo:epool-query 
							    ": ion pool component " (third x)
							    " must export a state"))
					    `(,@(cdr x) ,(car exports))))
					   epool-comps)))
		       (fold recur (append epools ax) (map third other-comps))))))))

(define (nemo:gate-complex-query sys . rest)
  (let-optionals rest ((ionic-current-name     (lambda (ion-name) (s+ 'i ion-name)))
		       (rev-potential-name     (lambda (ion-name) (s+ 'e ion-name )))
		       (in-concentration-name  (lambda (ion-name) (s+ ion-name 'i)))
		       (out-concentration-name (lambda (ion-name) (s+ ion-name 'o))))

  (match-let ((($ nemo:quantity 'DISPATCH  dis) (environment-ref sys (nemo-intern 'dispatch))))
    (let ((imports  ((dis 'imports)  sys))
	  (exports  ((dis 'exports)  sys)))
      (let* ((consts      ((dis 'consts)  sys))
	     (asgns       ((dis 'asgns)   sys))
	     (states      ((dis 'states)  sys))
	     (reactions   ((dis 'reactions) sys))
	     (rates       ((dis 'rates) sys))
	     (defuns      ((dis 'defuns)  sys))
	     (components  ((dis 'components) sys))
	     (gate-complexes     (filter-map (match-lambda 
	                                                   ((name 'gate-complex id) (list name id)) 
	                                                   ((name 'ion-channel id) (list name id)) 
	                                                   (else #f)) components))
	     (ion-pools   (nemo:ion-pool-query sys))
	
	     (perm-ions         (fold (lambda (gate-complex ax) 
					(let* ((subcomps ((dis 'component-subcomps) sys (cid gate-complex)))
					       (perm      (lookup-def 'permeating-ion subcomps)))
					  (cond (perm 
						 (case (cn perm)
						   ((non-specific)   
						    (let* ((erev (car ((dis 'component-exports) sys (cid perm))))
							   (i 'i) (e 'e))
						      (cons `(,(cn perm) ,i ,e ,erev) ax)))
						   (else (let* ((erev ((lambda (x) (and (pair? x) (car x))) ((dis 'component-exports) sys (cid perm))))
								(i    (ionic-current-name (cn perm)))
								(e    (rev-potential-name (cn perm))))
							   (cons `(,(cn perm) ,i ,e ,erev) ax)))))
						(else ax))))
				      (list) gate-complexes))
	     
	     (perm-ions         (fold (lambda (gate-complex ax) 
					(let* ((subcomps ((dis 'component-subcomps) sys (cid gate-complex)))
					       (bingate   (lookup-def 'binary-gate subcomps)))
					  (cond 
						(bingate 
						 (cons `(non-specific i e #f) ax))
						(else ax))))
				      perm-ions gate-complexes))
	     
	     (acc-ions           (fold (lambda (gate-complex ax) 
					 (let* ((subcomps ((dis 'component-subcomps) sys (cid gate-complex)))
						(acc   (lookup-def 'accumulating-substance subcomps))
						(i     (and acc (ionic-current-name (cn acc))))
						(in    (and acc (in-concentration-name (cn acc))))
						(out   (and acc (out-concentration-name (cn acc)))))
					   (if acc  (cons `(,(cn acc) ,i ,in ,out) ax) ax)))
				       (list) gate-complexes))

	     (pool-ions          (map (lambda (ep)
					(let* ((ion    (first ep))
					       (state  (third ep))
					       (i      (ionic-current-name ion))
					       (in     (in-concentration-name ion)))
					  `(,state ,i ,in ,ion)))
				      ion-pools))

	     (i-gates           (fold (lambda (gate-complex ax) 
					(let* ((subcomps     ((dis 'component-subcomps) sys (cid gate-complex)))
					       (i-gate-comp  (lookup-def 'binary-gate subcomps)))
					  (if i-gate-comp
					      (let* ((i-gate-exps  ((dis 'component-exports) sys (cid i-gate-comp)))
						     (asgn-names   asgns)
						     (i-gate-var   (find (lambda (x) (member x asgn-names)) 
									 i-gate-exps)))
						(if (not i-gate-var)
						    (nemo:error 'nemo:gate-complex-query 
								": binary gate in gate complex " (cn gate-complex)
								" must export an assignment"))
						(cons (list i-gate-var (cn i-gate-comp) ) ax))
					      ax)))
				      (list) gate-complexes)))

	   (for-each 
	    (lambda (a)
	      (let ((acc-ion   (car a)))
		(if (assoc acc-ion perm-ions)
		    (nemo:error 'nemo:gate-complex-query 
				": ion species " acc-ion " cannot be declared as both accumulating and permeating"))))
	    acc-ions)

	   (for-each 
	    (lambda (p)
	      (let ((pool-ion  (car p)))
		(if (assoc pool-ion perm-ions)
		    (nemo:error 'nemo:gate-complex-query 
				": ion species " pool-ion " cannot be declared as both pool and permeating"))))
	    pool-ions)
	
	`((gate-complexes ,gate-complexes)
	  (perm-ions    ,perm-ions)
	  (acc-ions     ,acc-ions)
	  (pool-ions    ,pool-ions)
	  (i-gates      ,i-gates)
	  ))))))
	  
)
